/*
 * Copyright (c)1997 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@enteract.com>.
 *
 * ----------------------------------------------------------------------
 *
 * ArpPTYRemote.cpp
 *
 */


#ifndef PTYREMOTE_H
#include "PTYRemote.h"
#endif

#ifndef PTYREMOTESETTINGS_H
#include "PTYRemoteSettings.h"
#endif

#ifndef ARPTERMINAL_ARPEMULATOR_H
#include <ArpTerminal/ArpEmulator.h>
#endif

#ifndef ARPTERMINAL_ARPTERMINALMSG_H
#include "ArpTerminal/ArpTerminalMsg.h"
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef _UTF8_H
#include <be/support/UTF8.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>

#include <OS.h>
#include <image.h>
#include <errno.h>
#include <stdlib.h>
#include <termios.h>

ArpMOD();

/*	------------------------------------------------------------
	PUBLIC CONFIGURATION INFORMATION.
	------------------------------------------------------------ */

const char* ArpPTYRemote::CommandConfigName = "Command";

/* ----------------------------------------------------------
	"ArpPTYRemote" class add-on interface.
   ---------------------------------------------------------- */
   
enum { NUM_TYPES = 1 };

static const ArpRemoteType type_info[] = {
	{ sizeof(ArpRemoteType), "shell", "Shell", NULL }
};

static int32 DoCountDevices(void)
{
	return NUM_TYPES;
}

static const ArpRemoteType* DoDeviceType(int32 idx)
{
	if( idx < 0 || idx >= NUM_TYPES ) return NULL;
	return &type_info[idx];
}

static ArpRemoteInterface* DoAllocRemote(const char* name)
{
	if( !name ) return NULL;
	if( !strcmp(name, type_info[0].URL) ) return new ArpPTYRemote;
	return NULL;
}

const ArpRemoteAddon ArpPTYRemote::AddonInterface = {
	ArpRemote_Current,
	
	"PTY Remote Device",
	"Angry Red Planet",
	B_UTF8_COPYRIGHT "1998 Dianne Hackborn and Angry Red Planet",
	__DATE__, __TIME__,
	1,
	
	"This remote device is an interface to the system's shell.  "
	"In use, it should work pretty much just like Be's Terminal "
	"application.  (Many thanks to Brian J. Swetland for the PTY "
	"code and ArpTerminal wrapper upon which this is based.)",
	
	&DoCountDevices,
	&DoDeviceType,
	&DoAllocRemote,
};

#define ctrl(c) ((c) - 0100)  
              
const static struct termios tdefault = {
	ICRNL,                          /* c_iflag */
	OPOST|ONLCR,                    /* c_oflag */
	B19200|CS8|CREAD|HUPCL,         /* c_cflag */
	ISIG|ICANON|ECHO|ECHOE|ECHONL,  /* c_lflag */
	0,                              /* c_line */
	0,                              /* c_ixxxxx */
	0,                              /* c_oxxxxx */
	{
		ctrl( 'C'),                 /* c_cc[VINTR] */
		ctrl( '\\'),                /* c_cc[VQUIT] */
		ctrl( 'H'),                 /* c_cc[VERASE] */
		ctrl( 'U'),                 /* c_cc[VKILL] */
		ctrl( 'D'),                 /* c_cc[VEOF] */
		0,                          /* c_cc[VEOL] */
		0,                          /* c_cc[VEOL2] */
		0,                          /* c_cc[VSWTCH] */
		ctrl( 'S'),                 /* c_cc[VSTART] */
		ctrl( 'Q'),                 /* c_cc[VSTOP] */
		ctrl( 'Z')                  /* c_cc[VSUSP] */
	}
};

          
/* find a tty/pty pair, launch argv[0], return ptyfd on success, -1 on failure */
static int32 ExecControlled(char *argv[], int argc, thread_id *thr)
{
	uint32 count;
	char *suf1Pt = "pqrstuvwxyzPQRST";
	//char ptyDev[20];
	char ttyDev[20];
	int32 ttyfd = -1;
	int32 ptyfd = -1;
	int32 pgid;
	thread_id shellThread;
	int stdinfd, stdoutfd, stderrfd;
		
	/* open a pseudoterminal */
	while(*suf1Pt && ttyfd < 0) {
		for(count = 0; count < 16; count++) {
			/* find an unused pty/tty name */
			char ptyDev[20];
			sprintf(ptyDev, "/dev/pt/%c%lx", *suf1Pt, count);
			sprintf(ttyDev, "/dev/tt/%c%lx", *suf1Pt, count);
			ptyfd = open(ptyDev, O_RDWR);
			if (ptyfd >= 0) {
				/* pty master opened, try opening matching slave */
				ttyfd = open(ttyDev, O_RDWR);
				if (ttyfd >= 0) {
					/* we are done */
					break;
				}				
				/* failed to open slave, try with another master */
				close(ptyfd);
				ptyfd = -1;
			} else if (errno == ENOENT) {	
				break;
			}				
		}
		suf1Pt++;		 
	}
	
	if (ptyfd < 0) {
		fprintf(stderr,"failed to open a pty\n");
		return -1;
	}
	
	/* initialize the tty */
	ioctl(ttyfd, TCSETA, &tdefault);
	
	/* save copies of our existing STDIO */
	stdinfd = dup(STDIN_FILENO);
	stdoutfd = dup(STDOUT_FILENO);
	stderrfd = dup(STDERR_FILENO);
	
	/* close the originals */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
	
	/* hook up the tty */
	dup(ttyfd);
	dup(ttyfd);
	dup(ttyfd);
	
	/* ensure the spawned process doesn't keep our fd's */
	fcntl(stdinfd, F_SETFD, FD_CLOEXEC);
	fcntl(stdoutfd, F_SETFD, FD_CLOEXEC);
	fcntl(stderrfd, F_SETFD, FD_CLOEXEC);
	fcntl(ptyfd, F_SETFD, FD_CLOEXEC);
	fcntl(ttyfd, F_SETFD, FD_CLOEXEC);
        
		/* prepare an environment */
	{
		char buffer[40];
		sprintf(buffer, "TTY=%s", ttyDev);
		putenv("TERM=beterm");
		putenv(buffer);
	}


	/* load the new process */
	shellThread = load_image(argc, (const char**)argv,
							 (const char**)environ);

	/* restore STDIO */
	dup2(stdinfd, STDIN_FILENO);
	dup2(stdoutfd, STDOUT_FILENO);
	dup2(stderrfd, STDERR_FILENO);
	close(stdinfd);
	close(stdoutfd);
	close(stderrfd);
	close(ttyfd);
	
	if (shellThread < 0) {             
		close(ptyfd);
		fprintf(stderr,"can't find the shell\n");
		return -1;
	}

	*thr = shellThread;
	
	/* start the child process goinf */
	setpgid(shellThread, shellThread);
	resume_thread(shellThread);
	pgid = shellThread;
	ioctl(ptyfd, 'pgid', pgid);
	
	return ptyfd;
}


ArpPTYRemote::ArpPTYRemote(const BMessenger& term, const char* url)
	: BLooper("ARP™ PTY Remote"), terminal(term),
	  mWatch(this, this)
{
	initialize(url);
}

ArpPTYRemote::ArpPTYRemote(const char* url)
	: BLooper("ARP™ PTY Remote"),
	  mWatch(this, this)
{
	initialize(url);
}

ArpPTYRemote::~ArpPTYRemote()
{
	DoDisconnect();
}

void ArpPTYRemote::initialize(const char* url)
{
	// Global state
	socketError = B_NO_ERROR;
	doEcho = true;
	curEmuType = -1;
	curEmuName = "";
	curWidth = curHeight = 0;
	
	// Default command line to execute
	cmdline = "/bin/sh -login";
	
	// No files currently open
	ptyfd = -1;
	
	// No shell is running
	shellThread = -1;
	connected = false;
	mInUse = false;
	
	// Reader thread and its associated private data.
	// (Note that the socket input handle is also private to
	//  the reader thread.)
	readThread = -1;
	mDisconnectRequested = false;
	
	// Start looper to receive input from terminal.
	Run();
	
	if( url ) Connect(url);
}

status_t ArpPTYRemote::GetConfiguration(BMessage* values) const
{
	if( !values ) return B_BAD_VALUE;
	
	BAutolock l(const_cast<ArpPTYRemote*>(this));
	if( !l.IsLocked() ) return B_ERROR;
	
	status_t res = values->AddString(CommandConfigName, cmdline.String());
	
	return res;
}

status_t ArpPTYRemote::PutConfiguration(const BMessage* values)
{
	status_t res = B_OK;
	
	if( !values ) return B_BAD_VALUE;
	
	BAutolock l(this);
	if( !l.IsLocked() ) return B_ERROR;
	
	if( res == B_OK && values->HasString(CommandConfigName) ) {
		const char* str;
		res = values->FindString(CommandConfigName, &str);
		if( res == B_OK ) cmdline = str;
	}
		
	if( res == B_OK ) mWatch.ReportChange(values);
	
	return res;
}

status_t ArpPTYRemote::Configure(ArpVectorI<BView*>& views)
{
	status_t err = B_NO_ERROR;
	
	BAutolock l(this);
	if( !l.IsLocked() ) return B_ERROR;
	
	BMessage settings;
	if( (err=GetConfiguration(&settings)) != B_NO_ERROR ) {
		ArpD(cdb << ADH << "Error getting settings: " << err << endl);
		return err;
	}
	
	views.push_back(new PTYRemoteSettings(BMessenger(this), settings));
	
	return err;
}

const ArpRemoteType* ArpPTYRemote::RemoteType(void) const
{
	return &type_info[0];
}
	
status_t ArpPTYRemote::Connect(BWindow* win)
{
	BAutolock llock(this);
	BAutolock glock(&globals);
	status_t res = B_ERROR;
	BMessage connmsg(TERM_CONNECT_MSG);
	res = PostMessage(&connmsg, this);
	return res;
}

void ArpPTYRemote::Disconnect()
{
	BAutolock llock(this);
	BAutolock glock(&globals);
	BMessage connmsg(TERM_DISCONNECT_MSG);
	PostMessage(&connmsg, this);
}

status_t ArpPTYRemote::Connect(const char* url, BWindow* win)
{
	BAutolock llock(this);
	BAutolock glock(&globals);
	status_t res = B_ERROR;
	if( url ) {
		BMessage connmsg(TERM_CONNECT_MSG);
		if( (res=connmsg.AddString("url", url)) == B_NO_ERROR ) {
			res = PostMessage(&connmsg, this);
		}
	} else res = Connect(win);
	
	return res;
}

status_t ArpPTYRemote::DoConnect(const char* url)
{
	BAutolock llock(this);
	
	DoDisconnect();
	
	BAutolock glock(&globals);
	
	socketError = B_ERROR;
	
	waitingReply = false;
	mDisconnectRequested = false;
	
	BString mungeline(cmdline);
	const int mungelen = mungeline.Length();
	char* pos = mungeline.LockBuffer(mungelen+1);
	
	char* argv[128];
	int32 argc=0;
	while( pos && *pos && argc < 127 ) {
		while( *pos && isspace(*pos) ) pos++;
		if( *pos == '"' ) {
			pos++;
			argv[argc++] = pos;
			char* put = pos;
			while( *pos && *pos != '"' ) {
				if( *pos == '\\' ) pos++;
				if( put < pos ) *put = *pos;
				put++;
				pos++;
			}
		} else {
			argv[argc++] = pos;
			while( *pos && !isspace(*pos) ) pos++;
		}
		
		if( *pos ) {
			*pos = 0;
			pos++;
		}
	}
	argv[argc] = 0;
	
	mungeline.UnlockBuffer(mungelen);
	
	if((ptyfd = ExecControlled(argv, argc, &shellThread)) >= 0) {
	
		// Now start reader thread to receive input from remote.
	
		BAutolock llock(this);
	
		BAutolock glock(&globals);
	
		// Just start up the reader thread, and let it do all else.
		ArpD(cdb << ADH << "ArpPTYRemote: Spawn reader thread." << endl);
		readThread = spawn_thread(readThreadEntry,
									"ARP™ PTY Reader",
									B_NORMAL_PRIORITY,
									this);
		if( readThread < 0 ) {
			Disconnect();
			return (status_t)readThread;
		}
		ArpD(cdb << ADH << "ArpPTYRemote: Resume reader thread." << endl);
		socketError = resume_thread(readThread);
		
		// Kill the thread if unable to start it.
		if( socketError ) {
			Disconnect();
			return socketError;
		}
		
		// Initialize pty to current dimensions of terminal.
		setWindowSize(curWidth, curHeight, true);
	}
	
	return socketError;
}

void ArpPTYRemote::DoDisconnect()
{
	BAutolock llock(this);
	
	ArpD(cdb << ADH << "ArpPTYRemote: Disconnecting..." << endl);

	globals.Lock();
	
	thread_id mythread = readThread;
	readThread = -1;
	mDisconnectRequested = true;
	
	if( shellThread >= 0 ) {
		send_signal(shellThread, SIGHUP);
		shellThread = -1;
	}
	
	// If the connected flag has been set and the thread is
	// running, it's in its read loop and needs to be forced
	// to quit.
	while( connected && mythread >= 0 ) {
		ArpD(cdb << ADH << "ArpPTYRemote: signal " << SIGUSR1
							<< "to reader " << mythread << endl);
		send_signal(mythread, SIGUSR1);
		globals.Unlock();
		ArpD(cdb << ADH << "ArpPTYRemote: waiting for reader." << endl);
		snooze(20000);
		//wait_for_thread(mythread,&ret);
		ArpD(cdb << ADH << "ArpPTYRemote: checking reader again." << endl);
	}
	ArpD(cdb << ADH << "ArpPTYRemote: no reader running." << endl);
	
	globals.Unlock();
	mInUse = false;
	mDisconnectRequested = false;
	
	if( ptyfd >= 0 ) {
		close(ptyfd);
		ptyfd = -1;
	}
}

/* ------------------------------------------------------------
   HIGH-LEVEL MESSAGE HANDLING
   ------------------------------------------------------------ */

void ArpPTYRemote::MessageReceived(BMessage* message)
{
	if( !message ) return;
	
	ArpD(cdb << ADH << "Telnet::MessageReceived: " <<
				*message << endl);

	if( mWatch.MessageReceived(message) == B_OK ) return;
	
	if( message->IsReply() ) {
		ArpD(cdb << ADH << "This is a reply." << endl);
		const BMessage* reply = message->Previous();
#if defined(ArpDEBUG)
		if( reply ) {
			ArpD(cdb << ADH << "Previous msg=" << *reply << endl);
		} else {
			ArpD(cdb << ADH << "No previous message!" << endl);
		}
#endif
		if( reply && reply->what == TERM_XFER_TEXT_MSG ) {
			ArpD(cdb << ADH << "Emptying out current buffer." << endl);
			BAutolock glock(&globals);
			waitingReply = false;
			receive("",0);
		}
	}
	
	switch( message->what ) {
		case TERM_CONNECT_MSG: {
			const char* url = NULL;
			if( message->HasString("url") ) {
				if( message->FindString("url", &url) != B_NO_ERROR ) {
					return;
				}
			}
			DoConnect(url);
		} break;
		case TERM_DISCONNECT_MSG: {
			DoDisconnect();
		} break;
		case TERM_XFER_TEXT_MSG: {
			const char* str = NULL;
			ssize_t len = 0;
			int32 pos = 0;
			while( message->FindData("text", B_ASCII_TYPE, pos,
									(const void**)&str, &len) == B_NO_ERROR ) {
				if( str && len > 0 ) write(str,len);
				pos++;
			}
		} break;
		case TERM_ATTACH_MSG: {
			message->FindMessenger("terminal",&terminal);
			emuTypes = BMessage();
		} break;
		case TERM_WINDOW_SIZE_MSG: {
			int32 width=0, height=0;
			if( message->FindInt32("columns",&width) == B_OK
				&& message->FindInt32("rows",&height) == B_OK ) {
				if( curWidth != width || curHeight != height ) {
					setWindowSize(width,height);
				}
			}
		} break;
		case TERM_EMULATORS_MSG: {
			emuTypes = *message;
			#if 0
			if( delayedTTYPE ) {
				send_ttype();
				delayedTTYPE = false;
			}
			#endif
		} break;
		case TERM_SELECT_EMULATION_MSG: {
			const char* name = NULL;
			if( message->FindString("name",&name) == B_OK ) {
				setEmulator(name);
			}
		} break;
		default:
			inherited::MessageReceived(message);
	}
}

void ArpPTYRemote::receive(const char* str, int32 len)
{
	if( !str ) return;
	if( len < 0 ) len = strlen(str);
	
	BAutolock glock(&globals);
	
	if( !terminal.IsValid() ) return;
	
	ArpD(cdb << ADH << "Receive text: " << str << endl);
	
	if( waitingReply ) {
		writeBuf.Append((const ichar*)str,len);
		ArpD(cdb << ADH << "Waiting reply.  New buffer is:" << endl
					<< "  " << writeBuf << endl);
		return;
	}
	
	BMessage msg(TERM_XFER_TEXT_MSG);
	
	if( !writeBuf.IsEmpty() ) {
		ArpD(cdb << ADH << "Appending to old buffer and sending." << endl);
		if( len > 0 ) writeBuf.Append((const ichar*)str,len);
		if( msg.AddData("text",B_ASCII_TYPE,(const char*)writeBuf,
					writeBuf.Length()) != B_NO_ERROR ) {
			return;
		}
	} else {
		if( len <= 0 ) return;
		ArpD(cdb << ADH << "Sending." << endl);
		if( msg.AddData("text",B_ASCII_TYPE,str,len) != B_NO_ERROR ) {
			return;
		}
	}
	
	// A quick and dirty way to buffer data going to terminal.
	terminal.SendMessage(&msg, this);
	writeBuf = "";
	waitingReply = true;
}

void ArpPTYRemote::changeEmulator(const char* name)
{
	BAutolock glock(&globals);
	
	if( !terminal.IsValid() ) return;
	
	if( !writeBuf.IsEmpty() ) {
		BMessage msg(TERM_XFER_TEXT_MSG);
		if( msg.AddData("text",B_ASCII_TYPE,(const char*)writeBuf,writeBuf.Length())
					== B_NO_ERROR ) {
			terminal.SendMessage(&msg, this);
			writeBuf = "";
			waitingReply = true;
		}
	}
	
	BMessage msg(TERM_SELECT_EMULATION_MSG);
	
	if( msg.AddString("name",name) != B_NO_ERROR ) return;
	
	terminal.SendMessage(&msg);
}

/* ------------------------------------------------------------
   DATA SENDING METHODS
   ------------------------------------------------------------ */

void ArpPTYRemote::setWindowSize(int32 width, int32 height, bool force)
{
	BAutolock glock(&globals);
	
	if( force || curWidth != width || curHeight != height ) {
		curWidth = width;
		curHeight = height;
		struct winsize ws;
		ws.ws_row = height;
		ws.ws_col = width;
		ws.ws_xpixel = width*8;
		ws.ws_ypixel = height*8;
		ioctl(ptyfd,TIOCSWINSZ,(char*)&ws);
		//send_signal(shellThread, SIGWINCH);
		//send_naws(width,height);
	}
}

void ArpPTYRemote::setEmulator(const char* name)
{
	BAutolock glock(&globals);
	
	if( name && curEmuName != (const ichar*)name ) {
		curEmuName = name;
		// TELNET protocol doesn't let us dynamically change, yes?
	}
}

void ArpPTYRemote::write(char b)
{
	BAutolock glock(&globals);
	
	if( connected ) {
		ArpD(cdb << ADH << "ArpPTYRemote: Put: " << (char)b
							<< " (" << (int)b << ")" << endl);
		int size = ::write(ptyfd, &b, 1);
		ArpD(cdb << ADH << "Bytes written: " << size << endl);
		ArpD(cdb << ADH << "Current error: " << strerror(errno) << endl);
		if( doEcho ) {
			receive(&b,1);
		}
		if( size < 1 ) {
			ArpString str("Error: ");
			str += strerror(errno);
			receive((const char*)str, str.Length());
		}
		mInUse = true;
	} else receive(&b,1);
}
  
void ArpPTYRemote::write(const char* b, int32 len)
{
	BAutolock glock(&globals);
	
	if( connected && len > 0 ) {
		ArpD(cdb << ADH << "WebTerm: Put: " << b
							<< " (" << len << " bytes)" << endl);
		int size = ::write(ptyfd, b, len);
		ArpD(cdb << ADH << "Bytes written: " << size << endl);
		ArpD(cdb << ADH << "Current error: " << strerror(errno) << endl);
		if( doEcho ) {
			receive(b,len);
		}
		if( size < len ) {
			ArpString str("Error: ");
			str += strerror(errno);
			receive((const char*)str, str.Length());
		}
		mInUse = true;
	} else if( len > 0 ) receive(b,len);
}
  
void ArpPTYRemote::write(const ArpString& str)
{
	write(str,str.Length());
}
  
/* ------------------------------------------------------------
   INPUT THREAD METHODS
   ------------------------------------------------------------ */

static void ignore_signal(int)
{
	ArpD(cdb << ADH << "ArpPTYRemote: entered ignore_signal(), thread="
						<< find_thread(NULL) << endl);
}

/* Entry point and main loop.
   Note well: we must be careful that any methods that read
   from the stream (either directly with in.read(), or indirectly
   with next_char()) are not synchronized.  Otherwise, they may
   block and leave other threads unable to access the class and
   write more data to the socket. */

int32 ArpPTYRemote::readThreadEntry(void* arg)
{
	ArpD(cdb << ADH << "ArpPTYRemote: Enter the reader." << endl);
	signal(SIGUSR1,ignore_signal);
	ArpPTYRemote *obj = (ArpPTYRemote *)arg; 
	return (obj->startReader()); 
}

int32 ArpPTYRemote::startReader(void)
{
	connected = true;
	doEcho = false;
	
	{
		BAutolock glock(globals);
		BMessage sessionMsg(TERM_START_SESSION_MSG);
		terminal.SendMessage(&sessionMsg);
	}
	
	int32 retval = runReader();

	{
		BAutolock glock(globals);
		BMessage sessionMsg(TERM_END_SESSION_MSG);
		sessionMsg.AddBool("remoteclosed",
						   mDisconnectRequested ? false : true);
		terminal.SendMessage(&sessionMsg);
	}
	
	doEcho = true;
	connected = false;
	mInUse = false;
	
	return retval;
}

int32 ArpPTYRemote::runReader(void)
{
	uint8 buffer[1024];
	
	//bool client_kill = false;
    ArpD(cdb << ADH << "WebTerm: Telnet input thread has started."
    				<< endl);
	ArpD(cdb << ADH << "WebTerm: Telnet thread: "
					<< find_thread(NULL) << endl);
	receive("Connected.\r\n");
	for(;;) {
		long len = 0;
		status_t ret = EINTR;
		if( readThread >= 0 ) {
			ArpD(cdb << ADH << "ArpPTYRemote: Ready to receive." << endl);
			ret = ::read(ptyfd, buffer, sizeof(buffer)-4);
		}
		//if( ret == 0 ) ret = ECONNABORTED;
		if( (ret < B_OK && ret != EINTR) || readThread < 0 ) {
			ArpD(cdb << ADH << "ArpPTYRemote: Input thread is terminating."
							<< endl);
			socketError = ret;
			return ret;
		}
		len = ret;
		if( len > 0 ) {
			ArpD(cdb << ADH << "ArpPTYRemote: Now read " << len << " bytes:"
							<< ArpEmulator::sequenceToString(ArpString(&buffer[0],0,len))
							<< endl);
			BAutolock llock(this);
			receive((const char*)&buffer[0],len);
		} else {
			ArpD(cdb << ADH << "ArpPTYRemote: Nothing read." << endl);
		}
	}
}

/* ------------------------------------------------------------
   MISC SUPPORT METHODS
   ------------------------------------------------------------ */

/* Parse escapes in a parameter string. */

ArpString ArpPTYRemote::parseString(const ArpString& in)
{
	if( ((const char*)in) == NULL || *((const char*)in) == 0 ) {
		return ArpString("");
	}
	ArpString out;
	for( int i=0; i<in.Length(); i++ ) {
		if( in[i] != '\\' ) out += (ichar)in[i];
		else if( ++i < in.Length() ) {
			switch( in[i] ) {
				case '\\':
					out += (ichar)'\\';
					break;
				case 'n':
					out += (ichar)'\n';
					break;
				case 'r':
					out += (ichar)'\r';
					break;
				case 't':
					out += (ichar)'\t';
					break;
				case 'f':
					out += (ichar)'\f';
					break;
				case 'b':
					out += (ichar)'\b';
					break;
				case 'x': {
					int c = 0;
					for( int j=0; j<2 && i<in.Length(); j++,i++ ) {
						c<<=4;
						c = c+(int)(in[i]-'0');
					}
					out += (ichar)c;
				} break;
				default: {
					if( in[i] >= '0' && in[i] <= '9' ) {
						int c = 0;
						for( int j=0; j<3 && i<in.Length(); j++,i++ ) {
							c<<=3;
							c = c+(int)(in[i]-'0');
						}
						out += (ichar)c;
					}
				} break;
			}
		}
	}
	return out;
}
